/* This file is part of Mailfromd.
   Copyright (C) 2006, 2007, 2008, 2009 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include "mailfromd.h"


#define BUCKETSIZE 16

struct symbucket {
	struct symbucket *next;
	size_t elsize;
	size_t index;
	char *buf;
};

struct _syment_align {
	char c;
	struct syment x;
};

#define __SYMENT_ALIGNMENT (offsetof(struct _syment_align, x))

#define __SYMENT_ALIGN(a, b) (((a) + (b) - 1) & ~((b) - 1))
#define SYMENT_SIZE(a) \
	__SYMENT_ALIGN((a)->elsize,__SYMENT_ALIGNMENT)
#define __SYMENT(a,p,n) \
	((struct syment*) ((char*) (p) + SYMENT_SIZE(a) * (n)))
#define SYMENT(a,n) \
	__SYMENT(a,(a)->buf,n)

static struct symbucket *symbucket;

/* Bucket handling */
static struct symbucket *
alloc_bucket(size_t elsize)
{
	size_t x;
	size_t size;
	struct symbucket *p;

	x = __SYMENT_ALIGN(elsize,__SYMENT_ALIGNMENT);
	size = ((sizeof(*p) + x - 1) / x) * x;
	p = xmalloc(size + x * BUCKETSIZE);
	p->index = 0;
	p->elsize = elsize;
	p->buf = (char*)p + x;
	p->next = symbucket;
	return symbucket = p;
}

static struct symbucket *
find_bucket(size_t elsize)
{
	struct symbucket *sp;
	for (sp = symbucket; sp; sp = sp->next)
		if (sp->elsize == elsize && sp->index < BUCKETSIZE)
			return sp;
	return alloc_bucket(elsize);
}

static void *
alloc_entry(size_t size)
{
	struct symbucket *bp = find_bucket(size);
	size_t n = bp->index++;
	return SYMENT(bp, n);
}

static void
free_buckets()
{
	struct symbucket *p = symbucket;
	while (p) {
		struct symbucket *next = p->next;
		free(p);
		p = next;
	}
	symbucket = NULL;
}


struct symtab *stab_builtin;
struct symtab *stab_function;
struct symtab *stab_variable;
struct symtab *stab_literal;
struct symtab *stab_constant;
struct symtab *stab_pragma;

struct symtab *
xsymtab_create(size_t elsize, int flags, void (*free_fun)(void *))
{
	struct symtab *s = symtab_create(elsize, flags, free_fun);
	if (!s)
		xalloc_die();
	return s;
}

struct syment *
xsymtab_lookup(struct symtab *st, const char *name)
{
	struct syment *ep;
	int rc = symtab_lookup_or_install(&ep, st, name, NULL);
	switch (rc) {
	case ENOMEM:
	case E2BIG:
		mu_error("%s", symtab_strerror(rc));
		abort();
	}
	return rc == 0 ? ep : NULL;
}
	
struct syment *
xsymtab_install(const char *what, struct symtab *st, const char *name)
{
	struct syment *ep;
	int install = 1;
	int rc = symtab_lookup_or_install(&ep, st, name, &install);
	if (rc) {
		parse_error(_("cannot install %s %s: %s"),
			    what, name, symtab_strerror(rc));
		abort();
	}
	return ep;
}


void
init_symbols()
{
	syment_alloc_hook = alloc_entry;
	stab_builtin = xsymtab_create(sizeof(struct builtin),
				      SYMTAB_COPY_KEY, NULL);
	stab_function = xsymtab_create(sizeof(struct function),
				       SYMTAB_COPY_KEY, NULL);
	stab_variable = xsymtab_create(sizeof(struct variable),
				       SYMTAB_COPY_KEY, NULL);
	stab_literal = xsymtab_create(sizeof(struct literal),
				      SYMTAB_COPY_KEY, NULL);
	stab_constant = xsymtab_create(sizeof(struct constant),
				       SYMTAB_COPY_KEY, NULL);
	stab_pragma = xsymtab_create(sizeof(struct pragma),
				     SYMTAB_COPY_KEY, NULL);
}

void
free_symbols()
{
	symtab_free(&stab_builtin);
	symtab_free(&stab_function);
	symtab_free(&stab_variable);
	symtab_free(&stab_literal);
	symtab_free(&stab_constant);
	symtab_free(&stab_pragma);

	free_buckets();
}



void
va_builtin_install(char *name,
		   void (*handler) (eval_environ_t),
		   data_type_t rettype,
		   size_t argcount,
		   ...)
{
	struct builtin *builtin;
	va_list ap;
	size_t i;
	data_type_t *argtype = xmalloc (argcount * sizeof *argtype);
	
	va_start(ap, argcount);
	for (i = 0; i < argcount; i++) 
		argtype[i] = va_arg(ap, data_type_t);
	va_end(ap);

	builtin = (struct builtin*) xsymtab_install(_("builtin function"),
						    stab_builtin, name);

	/* new entry */
	builtin->name = mf_strdup(name);
	builtin->handler = handler;
	builtin->parmcount = argcount;
	builtin->parmtype = argtype;
	builtin->rettype = rettype;
	builtin->statemask = 0;
	builtin->flags = 0;
}

void
va_builtin_install_ex (char *name,
		       void (*handler) (eval_environ_t),
		       unsigned statemsk,
		       data_type_t rettype,
		       size_t argcount,
		       size_t optcount,
		       int flags,
		       ...)
{
	struct builtin *builtin;
	va_list ap;
	size_t i;
	data_type_t *argtype = xmalloc (argcount * sizeof *argtype);

 	va_start(ap, flags);
	for (i = 0; i < argcount; i++) 
		argtype[i] = va_arg(ap, data_type_t);
	va_end(ap);

	builtin = (struct builtin*) xsymtab_install(_("builtin function"),
						    stab_builtin, name);

	/* new entry */
	builtin->name = mf_strdup(name);
	builtin->handler = handler;
	builtin->parmcount = argcount;
	builtin->optcount = optcount;
	builtin->parmtype = argtype;
	builtin->rettype = rettype;
	builtin->statemask = statemsk;
	builtin->flags = flags;
}

const struct builtin *
builtin_lookup(const char *name)
{
	return (struct builtin *)xsymtab_lookup(stab_builtin, name);
}



static void
init_variable(struct variable *var)
{
	var->type = dtype_unspecified;
	var->storage_class = storage_extern;
	var->off = 0;
	var->addrptr = NULL;
	var->shadowed = NULL;
	var->flags = 0;
	var->xref = NULL;
	var->next = NULL;
}	

struct variable *
variable_install(const char *name)
{
	struct variable *vp;
	struct syment *ent;
	int install = 1;
	int rc = symtab_lookup_or_install(&ent, stab_variable, name, &install);
	if (rc) {
		parse_error(_("cannot install %s %s: %s"),
			    _("variable"), name, symtab_strerror(rc));
		abort();
	}

	vp = (struct variable *) ent;
	if (install) {
		init_variable(vp);
		vp->name = mf_strdup(name);
	}
	return vp;
}

struct variable *
variable_replace(const char *name, struct variable *var)
{
	if (!var) {
		var = alloc_entry(sizeof(*var));
		init_variable(var);
		var->name = mf_strdup(name);
	}
	if (symtab_replace(stab_variable, (struct syment*)var, NULL)) {
		parse_error(_("INTERNAL ERROR: variable %s not found"),
			    name);
		abort();
	}
	return var;
}

struct variable *
variable_lookup(const char *name)
{
	return (struct variable *)xsymtab_lookup(stab_variable, name);
}


static struct function *
resolve_alias(struct function *fp)
{
	while (fp->alias) 
		fp = fp->alias;
	return fp;
}

static struct function *
function_lookup_or_install(const char *name, const struct locus *locus)
{
	struct function *fp;
	struct syment *ent;
	int install = 1;
	int rc = symtab_lookup_or_install(&ent, stab_function, name, &install);
	if (rc) {
		parse_error(_("cannot install %s %s: %s"),
			    _("function"), name, symtab_strerror(rc));
		abort();
	}

	fp = (struct function*) ent;
	if (!install) {
		if (fp->alias) {
			fp = resolve_alias(fp); /* FIXME: is it right? */
			parse_error_locus(locus,
				  _("redefinition of function %s by alias %s"),
					  fp->name, name);
			parse_error_locus(&fp->locus, 
		                  _("previously defined here"));
			/*FIXME: memory leak*/
		} else {
			parse_error_locus(locus,
					  _("redefinition of function %s"),
					  name);
			parse_error_locus(&fp->locus, 
					  _("previously defined here"));
			/*FIXME: memory leak*/
		}
	}
	return fp;
}

struct function *
function_install(const char *name, size_t parmcnt, size_t optcnt, int varargs,
		 data_type_t *parmtypes, data_type_t rettype,
		 const struct locus *locus)
{
	struct function *fp = function_lookup_or_install(name, locus);
	memset(fp, 0, sizeof(*fp));
	fp->name = mf_strdup(name);
	fp->locus = *locus;
	fp->node = NULL;
	fp->entry = 0;
	fp->parmcount = parmcnt;
	fp->optcount = optcnt;
	fp->varargs = varargs;
	fp->parmtype = parmtypes;
	fp->rettype = rettype;
	fp->exmask = 0;
	fp->statemask = 0;
	return fp;
}

struct function *
function_lookup(const char *name)
{
	struct function *fp =
		(struct function *)xsymtab_lookup(stab_function, name);
	if (fp)
		fp = resolve_alias(fp);
	return fp;
}


void
install_alias(const char *name, struct function *fun,
	      const struct locus *locus)
{
	struct function *fp = function_lookup_or_install(name, locus);
	memset(fp, 0, sizeof(*fp));
	fp->name = mf_strdup(name);
	fp->locus = *locus;
	fp->alias = fun;
}


struct literal *
literal_lookup(const char *text)
{
	struct literal *lit;
	struct syment *ent;
	int install = 1;
	int rc = symtab_lookup_or_install(&ent, stab_literal, text, &install);
	if (rc) {
		parse_error(_("cannot install %s %s: %s"),
			    _("literal"), text, symtab_strerror(rc));
		abort();
	}

	lit = (struct literal *)ent;
	lit->text = text; /*FIXME: already done in symtab_lookup_or_install */
	lit->flags = 0;
	lit->regex = NULL;
	return lit;
}


void
define_constant(const char *name, struct value *value, struct locus *locus)
{
	struct constant *cp;
	struct syment *ent;
	int install = 1;
	int rc = symtab_lookup_or_install(&ent, stab_constant, name, &install);
	if (rc) {
		parse_error(_("cannot install %s %s: %s"),
			    _("constant"), name, symtab_strerror(rc));
		abort();
	}

	cp = (struct constant *) ent;
	if (install) {
		cp->name = mf_strdup(name);
	} else {
		parse_warning_locus(locus,
				    "%s: redefinition of constant",
				    name);
		parse_warning_locus(&cp->locus,
				    "this is the location of the "
				    "previous definition");
	}
	cp->locus = *locus;
	cp->value = *value;
}

const struct constant *
constant_lookup(const char *name)
{
	return (const struct constant *)xsymtab_lookup(stab_constant, name);
}

const struct value *
constant_lookup_value(const char *name)
{
	const struct constant *cptr = constant_lookup(name);
	return cptr ? &cptr->value : NULL;
}


struct sym_regex *
install_regex(struct literal *lit, unsigned regflags)
{
	struct sym_regex *regex;

	for (regex = lit->regex; regex; regex = regex->next)
		if (regex->regflags == regflags)
			return regex;

	regex = alloc_entry(sizeof(*regex));
	regex->next = lit->regex;
	lit->regex = regex;
	regex->lit = lit;
	regex->regflags = regflags;
	regex->flags = 0;
	regex->index = 0;
	
	return regex;
}


void
install_pragma(const char *name, int minargs, int maxargs,
	       void (*handler) (int, char **, const char *))
{
	struct pragma *pragma;
	struct syment *ent;
	int install = 1;
	int rc = symtab_lookup_or_install(&ent, stab_pragma, name, &install);
	if (rc) {
		parse_error(_("cannot install %s %s: %s"),
			    _("pragma"), name, symtab_strerror(rc));
		abort();
	}
	if (!install) {
		parse_error(_("INTERNAL ERROR: pragma %s already defined"),
			    name);
		abort();
	}

	pragma = (struct pragma *)ent;
	pragma->name = mf_strdup(name);
	pragma->minargs = minargs;
	pragma->maxargs = maxargs;
	pragma->handler = handler;
}

const struct pragma *
lookup_pragma(const char *name)
{
	return (const struct pragma *)xsymtab_lookup(stab_pragma, name);
}
